<?php namespace Notes\Route;

use Notes\ObjectReflection,
    Notes\ObjectResolver;

use RuntimeException, DirectoryIterator, Generator, Closure;

class RouteFetcher {

    public array $list = [];

    protected bool $debug;

    protected array $folderList;

    protected Closure $callback;

    public array $defaultMethods = [ 'GET', 'POST' ];

    protected array $annotations;

    public function __construct(?Closure $callback = null, ? array $folderList = null, ? array $annotations = null, bool $debug = false)
    {
        // Handles if we must fetch or store in cache
        $this->debug = $debug;

        if ($callback !== null) {
            $this->callback = $callback;
        }

        if ($folderList !== null) {
            $this->folderList = $folderList;
        }

        if ($annotations !== null) {
            $this->annotations = $annotations;
        }
        else {
            $this->annotations = [
                'object' => Annotation\Object\Route::class,
                'method' => Annotation\Method\Route::class,
            ];
        }
    }

    public function __destruct() {
        if ($this->debug) {
            foreach ($this->list as $key => $item) {
                $cacheKey = sprintf("%s:routes.%s", basename(__CLASS__), $key);

                if (function_exists('apcu_enabled') && apcu_enabled()) {
                    apcu_store($cacheKey, $item, 3600);
                }
            }
        }
    }

    public function addFolder($folder) : void
    {
        $this->folderList[] = $folder;
    }

    public function setFolderList(array $list) : void
    {
        $this->folderList = $list;
    }

    public function scan(? array $folders = null) : Generator
    {
        foreach($folders ?: $this->folderList as $namespace => $folder) {
            if ( ! file_exists($folder) ) {
                throw new RuntimeException(sprintf("Folder `%s` can not be found or scanned", $folder));
            }

            foreach (new DirectoryIterator($folder) as $fileinfo) {
                if ( ! $fileinfo->isDot() ) {
                    if ( $fileinfo->isDir() ) {
                        foreach($this->scan([ "{$namespace}\\" . $fileinfo->getBasename() => $fileinfo->getPathname() ]) as $ns2 => $fi2) {
                            yield $ns2 => $fi2;
                        }
                    }
                    else {
                        yield $namespace => $fileinfo;
                    }
                }
            }
        }
    }

    public function compile(null|array $annotations = null) : array
    {
        $annotations ??= $this->annotations;

        if ( $cacheValue = $this->fetchFromCache($this->generateListKey($annotations)) ) {
            return $cacheValue;
        }

        foreach($this->scan() as $namespace => $file) {
            if (  $file->getExtension() !== "php" ) {
                continue;
            }

            $base = "";
            $class = $this->generateClassname($file->getBasename(".php"), $namespace);
            $methods = $this->defaultMethods;

            $objectResolver = new ObjectResolver($class, true, true, false, true);

            if ( isset($annotations['object']) && (null !== ( $object = $objectResolver->getAnnotationFromClassname($annotations['object']) )) ){
                if ( $object->methods ?? false ) {
                    $methods = $object->methods;
                }

                $base = $object->base ?? "";
            }

            if ( isset($annotations['method']) ) {
                $routeList = $objectResolver->getAnnotationListFromClassname( $annotations['method'] );

                foreach($routeList as $func => $routes) {
                    foreach($routes as $route) {
                        $route->base = $base;
                        $route->class = $class;
                        $route->classMethod = $func;

                        if (false === ($route->methods ?? false)) {
                            $route->methods = $methods;
                        }

                        if (false !== ($this->callback ?? false)) {
                            call_user_func_array($this->callback, [ $route ]);
                        }

                        $list[] = $route;
                    }
                }
            }
        }

        return $this->debug ? $this->list[$this->generateListKey($annotations)] = $list : $list;
    }

    protected function fetchFromCache(string $key) : array|null
    {
        if (! $this->debug) {
            if (!isset($this->list[$key]) && function_exists('apcu_enabled') && apcu_enabled()) {
                $cacheKey = sprintf("%s:routes.%s", basename(__CLASS__), $key);

                $fetch = apcu_fetch($cacheKey, $success);

                if ($success) {
                    return $fetch;
                }
            }
        }

        return $this->list[$key] ?? null;
    }

    protected function generateListKey(array $annotations) {
        return ($annotations['object'] ?? "") . ($annotations['method'] ?? "");
    }

    protected function generateClassname($file, $namespace)
    {
        return "\\$namespace\\$file";
    }
}
